<?php

namespace Verifone\Hosted\Helper;

use VerifoneEcomAPI\ApiWrapper\Http\ApiClient\Client;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Schemas\CaptureSchema;
use VerifoneEcomAPI\ApiWrapper\Schemas\CheckoutSchema;
use VerifoneEcomAPI\ApiWrapper\Schemas\CustomerSchema;
use VerifoneEcomAPI\ApiWrapper\Schemas\RefundSchema;
use VerifoneEcomAPI\ApiWrapper\Authentication\BasicAuth;
use VerifoneEcomAPI\ApiWrapper\Regions\Emea;
use VerifoneEcomAPI\ApiWrapper\Regions\Na;
use VerifoneEcomAPI\ApiWrapper\Regions\NzAu;
use VerifoneEcomAPI\ApiWrapper\Settings;

use Magento\Sales\Model\Order;
use Magento\Framework\App\CacheInterface;
use Exception;


class Verifone extends \Magento\Framework\App\Helper\AbstractHelper {
	const SUPPORTED_I18N_LANGUAGES = [ 'en', 'da', 'fi', 'fr', 'he', 'no', 'pl', 'pt', 'sv' ];
	const NORWAY_CODES = [ 'nb', 'nn' ];
	const VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_AUTHORIZED = 'AUTHORIZED';
	const VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_SETTLEMENT_COMPLETED = 'SETTLEMENT_COMPLETED';
	const VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_VOIDED = 'VOIDED';

	protected $_session;
	protected $_logger;

	/**
	 * @var \Magento\Sales\Model\OrderFactory
	 */
	protected $_orderFactory;

	/**
	 * @var \Magento\Framework\Api\SearchCriteriaBuilder
	 */
	protected $_searchCriteriaBuilder;

	/**
	 * @var \Magento\Quote\Model\QuoteRepository
	 */
	protected $_quoteRepo;

	/**
	 * @var CacheInterface
	 */
	private $_cache;

	/**
	 * @var \Magento\Sales\Model\Service\InvoiceService
	 */
	protected $_invoiceService;

	/**
	 * @var \Magento\Framework\DB\TransactionFactory
	 */
	protected $_transactionFactory;

	protected $_transactionBuilder;

	private $_verifoneSettings;
	private $_simpleCurl;
	private $_mapper;
	private $_client;


	public function __construct(
		\Magento\Framework\App\Helper\Context $context,
		\Magento\Checkout\Model\Session $session,
		\Magento\Framework\App\CacheInterface $cache,
		\Magento\Sales\Model\OrderFactory $orderFactory,
		\Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder,
		\Magento\Quote\Model\QuoteRepository $quoteRepo,
		\Magento\Sales\Model\Service\InvoiceService $invoiceService,
		\Magento\Framework\DB\TransactionFactory $transactionFactory,
		\Magento\Sales\Model\Order\Payment\Transaction\BuilderInterface $transactionBuilder
	) {
		$this->_session               = $session;
		$this->_cache                 = $cache;
		$this->_logger                = $context->getLogger();
		$this->_orderFactory          = $orderFactory;
		$this->_searchCriteriaBuilder = $searchCriteriaBuilder;
		$this->_quoteRepo             = $quoteRepo;
		$this->_invoiceService        = $invoiceService;
		$this->_transactionFactory    = $transactionFactory;
		$this->_transactionBuilder    = $transactionBuilder;
		parent::__construct( $context );

		$this->_simpleCurl = new SimpleCurl();
	}

	/**
	 * @param $comment
	 *
	 * @return bool
	 * @throws \Magento\Framework\Exception\LocalizedException
	 */
	public function cancelCurrentOrder( $comment ) {
		$order = $this->_session->getLastRealOrder();
		if ( $order->getId() && $order->getState() != \Magento\Sales\Model\Order::STATE_CANCELED ) {
			$order->registerCancellation( $comment )->save();

			return true;
		}

		return false;
	}

	/**
	 * @return bool
	 */
	public function restoreQuote() {
		return $this->_session->restoreQuote();
	}

	/**
	 * @param       $route
	 * @param array $params
	 *
	 * @return string
	 */
	public function getUrl( $route, $params = [] ) {
		return $this->_getUrl( $route, $params );
	}

	/**
	 * @param $incrementId
	 * @param $refNo
	 *
	 * @return false|\Magento\Sales\Model\Order
	 */
	public function getOrderByIncrementId( $incrementId, $refNo ) {
		$foundOrder = $this->_orderFactory->create()->loadByIncrementId( $incrementId );
		if ( $foundOrder->getId() && $refNo == $foundOrder->getExtOrderId() ) {
			return $foundOrder;
		}

		return false;
	}

	/**
	 * @param $incrementId
	 *
	 * @return \Magento\Quote\Api\Data\CartInterface|null
	 */
	public function getQuoteByIncrementId( $incrementId ) {
		$sc       = $this->_searchCriteriaBuilder->addFilter( \Magento\Quote\Api\Data\CartInterface::KEY_RESERVED_ORDER_ID,
			$incrementId )->create();
		$quoteArr = $this->_quoteRepo->getList( $sc )->getItems();
		$quote    = null;

		if ( count( $quoteArr ) > 0 ) {
			$quote = current( $quoteArr );
		}

		return $quote;
	}

	/**
	 * @param $paymentMethod
	 */
	public function setUpVerifoneSettings( $paymentMethod ) {
		$this->_verifoneSettings = new Settings();
		$this->_verifoneSettings->setTest( '1' == $paymentMethod->getConfigData( 'demo_mode' ) )
		                        ->setRegion( $paymentMethod->getConfigData( 'region' ) )
		                        ->setApiUserId( $paymentMethod->getConfigData( 'api_user' ) )
		                        ->setApiKey( $paymentMethod->getConfigData( 'api_key' ) )
		                        ->setTestApiUserId( $paymentMethod->getConfigData( 'test_api_user' ) )
		                        ->setTestApiKey( $paymentMethod->getConfigData( 'test_api_key' ) )
		                        ->set3dsEnable( '1' == $paymentMethod->getConfigData( 'tds' ) )
		                        ->set3dsContractId( $paymentMethod->getConfigData( 'tds_contract_id' ) )
		                        ->setPaymentContractId( $paymentMethod->getConfigData( 'payment_contract_id' ) )
		                        ->setEntityId( $paymentMethod->getConfigData( 'entity_id' ) )
		                        ->setTestEntityId( $paymentMethod->getConfigData( 'test_entity_id' ) )
		                        ->setTestPaymentContractId( $paymentMethod->getConfigData( 'test_payment_contract_id' ) )
		                        ->setTest3dsContractId( $paymentMethod->getConfigData( 'test_tds_contract_id' ) )
		                        ->setCurlVerifySsl( '1' == $paymentMethod->getConfigData( 'verifone_curl' ) )
		                        ->setCvvOnPayment( '1' == $paymentMethod->getConfigData( 'cvv_on_payment' ) )
		                        ->setPaymentType( Settings::PAYMENT_TYPE_HPP == $paymentMethod->getConfigData( 'payment_type' )
			                        ? Settings::PAYMENT_TYPE_HPP
			                        : Settings::PAYMENT_TYPE_IFRAME
		                        )
		                        ->setApplePay3dsEnable( '1' == strtolower( $paymentMethod->getConfigData( 'apple_pay_3ds_enable' ) ) )
		                        ->setGooglePay3dsEnable( '1' == strtolower( $paymentMethod->getConfigData( 'google_pay_3ds_enable' ) ) )
		                        ->setApplePayScaComplianceLevel( $paymentMethod->getConfigData( 'apple_pay_sca_compliance_level' ) )
		                        ->setTestApplePayThreedsContractId( $paymentMethod->getConfigData( 'test_apple_pay_threeds_contract_id' ) )
		                        ->setApplePayThreedsContractId( $paymentMethod->getConfigData( 'apple_pay_threeds_contract_id' ) )
		                        ->setApplePayContractId( $paymentMethod->getConfigData( 'apple_pay_threeds_contract_id' ) )
		                        ->setPaypalContractId( $paymentMethod->getConfigData( 'paypal_contract_id' ) )
		                        ->setGooglePayContractId( $paymentMethod->getConfigData( 'google_pay_contract_id' ) )
		                        ->setApplePayContractId( $paymentMethod->getConfigData( 'apple_pay_contract_id' ) )
		                        ->setTestApplePayThreedsContractId( $paymentMethod->getConfigData( 'test_google_pay_threeds_contract_id' ) )
		                        ->setApplePayThreedsContractId( $paymentMethod->getConfigData( 'google_pay_threeds_contract_id' ) )
		                        ->setTestPaypalContractId( $paymentMethod->getConfigData( 'test_paypal_contract_id' ) )
		                        ->setTestApplePayContractId( $paymentMethod->getConfigData( 'test_apple_pay_contract_id' ) )
		                        ->setTestGooglePayContractId( $paymentMethod->getConfigData( 'test_google_pay_contract_id' ) )
		                        ->setGooglePayScaComplianceLevel( $paymentMethod->getConfigData( 'google_pay_sca_compliance_level' ) )
		                        ->setIsApplePayEnabled( '1' == strtolower( $paymentMethod->getConfigData( 'apple_pay_enabled' ) ) )
		                        ->setIsGooglePayEnabled( '1' == strtolower( $paymentMethod->getConfigData( 'google_pay_enabled' ) ) )
		                        ->setIsPaypalEnabled( '1' == strtolower( $paymentMethod->getConfigData( 'paypal_enabled' ) ) )
		                        ->setManualCapture( '1' == $paymentMethod->getConfigData( 'capture' ) );
	}

	/**
	 * @param $paymentMethod
	 */
	public function getRegionMapper( $paymentMethod ) {
		// mini factory :D
		switch ( $paymentMethod->getConfigData( 'region' ) ) {
			case 'NZAU':
				$this->_mapper = new Nzau( $this->_verifoneSettings );
				break;
			case 'NA':
				$this->_mapper = new Na( $this->_verifoneSettings );
				break;
			case 'EMEA':
				$this->_mapper = new Emea( $this->_verifoneSettings );
				break;
			default:
				// will happen only on the first installation
				$this->_mapper = new Nzau( $this->_verifoneSettings );
		}
	}

	/**
	 * @param $order
	 * @param $payment
	 * @param $transactionId
	 * @param $paymentMethod
	 * @param string $type
	 *
	 * @return array|string[]|void
	 * @throws \VerifoneEcomAPI\ApiWrapper\Http\HttpException
	 */
	public function voidOrCapture( $order, $payment, $transactionId, $paymentMethod, string $type = 'capture' ) {
		$this->setUpVerifoneSettings( $paymentMethod );
		$this->getRegionMapper( $paymentMethod );
		$auth          = new BasicAuth( $this->_verifoneSettings );
		$this->_client = new Client(
			$this->_mapper,
			$this->_verifoneSettings,
			$auth,
			$this->_simpleCurl
		);

		$amount = $order->getGrandTotal() * 100;
		try {
			$payload = [
				'amount' => (int) str_replace( [ ',', '.' ], '', $amount ),
			];
			if ( $type == 'void' ) {
				$result = $this->_client->postVoidAuthorization( $payload, $transactionId );
			}
			if ( $type == 'capture' ) {
				$result = $this->_client->postCapture( $payload, $transactionId, new CaptureSchema() );
			}

		} catch ( \VerifoneEcommApi\ApiWrapper\Http\HttpException $exception ) {
			$this->_logger->critical( sprintf( 'There was an error in the void capture, last known error was "%s", in file %s, at line %s. Error message: %s',
				json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

			return $this->failure();
		}

		if ( isset( $result['code'] ) && $result['code'] == 400 ) {
			return;
		}
		if ( isset( $result['status'] ) && $result['status'] == self::VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_AUTHORIZED ) {
			$this->createTransactionForCapturePayment( $order, $payment, $transactionId, $result );
		}
		if ( isset( $result['status'] ) && $result['status'] == self::VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_SETTLEMENT_COMPLETED ) {
			$this->createTransactionForCapturePayment( $order, $payment, $transactionId, $result );
		}
		if ( isset( $result['status'] ) && $result['status'] == self::VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_VOIDED ) {
			$this->createTransactionForVoid( $order, $payment, $transactionId, $result );
		}
	}

	/**
	 * @return mixed
	 */
	public function checkManualCapture() {
		return $this->_verifoneSettings->getManualCapture();
	}

	/**
	 * @param $paymentMethod
	 * @param $params
	 *
	 * @return mixed
	 * @throws Exception
	 */
	public function checkPaymentResponse( $paymentMethod, $params ) {
		if ( ! isset( $params['checkoutId'] ) && ! isset( $params['checkout_id'] ) ) {
			$this->_logger->critical( 'Error message', [
				'exception' => sprintf( 'The checkoutId parameter could not be found, in order to proceed it must exist. Current GET params: %s',
					json_encode( $params ) )
			] );

			return false;
		}
		$checkoutId = ( isset( $params['checkoutId'] ) ) ? $params['checkoutId'] : $params['checkout_id'];
		if ( ( ! isset( $params['transactionId'] ) || $params['transactionId'] === '' || isset( $params['errorCode'] ) ) &&
		     ( ! isset( $params['transaction_id'] ) || $params['transaction_id'] === '' || isset( $params['error_code'] ) ) ) {
			$this->_logger->critical( 'Error message', [
				'exception' => sprintf( 'There was an error with the payment. Current GET params: %s',
					json_encode( $params ) )
			] );

			return false;
		}

		$this->setUpVerifoneSettings( $paymentMethod );
		$this->getRegionMapper( $paymentMethod );
		$auth          = new BasicAuth( $this->_verifoneSettings );
		$this->_client = new Client(
			$this->_mapper,
			$this->_verifoneSettings,
			$auth,
			$this->_simpleCurl
		);

		try {
			$response = $this->_client->getCheckout( $checkoutId );
		} catch ( \VerifoneEcommApi\ApiWrapper\Http\HttpException $exception ) {
			$this->_logger->critical( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
				json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

			return $this->failure();
		}

		if ( ! isset( $response['events'] ) ||
		     ( ! isset( $response['merchantReference'] ) && ! isset( $response['merchant_reference'] ) ) ||
		     ( ! isset( $response['transactionId'] ) && ! isset( $response['transaction_id'] ) ) ) {
			$this->_logger->critical( 'Error message', [
				'exception' => sprintf( 'The events or merchantReference or transactionId parameter could not be found in response, in order to proceed it must exist. API response %s',
					json_encode( $response ) )
			] );
			throw new Exception( 'An unexpected error occurred.' );
		}

		return $response;
	}

	/**
	 * @param $paymentMethod
	 * @param $quote
	 *
	 * @return array
	 * @throws Exception
	 */
	public function processPayment( $paymentMethod, $quote ) {
		$this->setUpVerifoneSettings( $paymentMethod );
		$this->getRegionMapper( $paymentMethod );
		$auth          = new BasicAuth( $this->_verifoneSettings );
		$this->_client = new Client(
			$this->_mapper,
			$this->_verifoneSettings,
			$auth,
			$this->_simpleCurl
		);

		if ( $this->_verifoneSettings->is3dsEnabled() && null === $this->_verifoneSettings->get3dsContractId() ) {
			$this->_logger->critical( 'Error message', [ 'exception' => '3DS is enabled but no 3DS contract ID is provided in the configuration. Please fill in the field "3DS Contract Id" or disable the field "Enable/Disable 3d-secure" in the plugin settings.' ] );
			throw new Exception( 'An unexpected error occurred. Please retry the payment. No money has been charged.' );
		}

		$billingAddress  = $quote->getBillingAddress();
		$shippingAddress = $quote->getShippingAddress();
		$billingArray    = $this->getBillingArray( $billingAddress );
		$shippingArray   = $this->getShippingArray( $shippingAddress, $billingAddress );

		$customerInfo = [
			'email_address' => substr( $billingAddress->getEmail(), 0, 250 ),
			'phone_number'  => preg_replace( '/[^0-9]+/', '', substr( $billingAddress->getTelephone(), 0, 20 ) ),
			'entity_id'     => $this->_verifoneSettings->getEntityId()
		];
		if ( ! empty( $billingAddress->getCompany() ) ) {
			$customerInfo['company_name'] = substr( $billingAddress->getCompany(), 0, 100 );
		}

		$customerInfo = array_merge( $customerInfo, $billingArray );
		$customerInfo = array_merge( $customerInfo, $shippingArray );

		try {
			$customer = $this->_client->postCustomer( $customerInfo );
		} catch ( \VerifoneEcommApi\ApiWrapper\Http\HttpException $exception ) {
			$this->_logger->critical( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
				json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

			return $this->failure();
		}
		$customerId = ( ! empty( $customer['id'] ) ) ? $customer['id'] : null;

		$currency      = $quote->getCurrency()->getQuoteCurrencyCode();
		$objectManager = \Magento\Framework\App\ObjectManager::getInstance();
		$getLocale     = $objectManager->get( 'Magento\Framework\Locale\Resolver' );
		$locale        = $getLocale->getLocale();
		$checkout      = $this->getCheckoutParams( $quote, $paymentMethod, $currency, $customerId, $locale );

		try {
			$checkout = $this->_client->postCheckout( $checkout );
		} catch ( \VerifoneEcommApi\ApiWrapper\Http\HttpException $exception ) {
			$this->_logger->critical( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
				json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

			return $this->failure();
		}
		$checkoutUrl = ( ! empty( $checkout['url'] ) ) ? $checkout['url'] : null;

		if ( $this->_verifoneSettings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
			$scheme = parse_url( $checkout['url'], PHP_URL_SCHEME );
			$host   = parse_url( $checkout['url'], PHP_URL_HOST );
			$url    = sprintf( '%s://%s/v2/loader.js?checkoutId=%s', $scheme, $host, $checkout['id'] );

			return [
				'method' => $this->_verifoneSettings->getPaymentType(),
				'url'    => $url
			];
		}

		return [
			'method'   => $this->_verifoneSettings->getPaymentType(),
			'redirect' => $checkoutUrl,
		];
	}

	public function processRefund( $paymentMethod, $amount, $reason, $order, $payment ) {
		$this->setUpVerifoneSettings( $paymentMethod );
		$this->getRegionMapper( $paymentMethod );
		$auth                  = new BasicAuth( $this->_verifoneSettings );
		$this->_client         = new Client(
			$this->_mapper,
			$this->_verifoneSettings,
			$auth,
			$this->_simpleCurl
		);
		$paymentAdditionalInfo = $payment->getAdditionalInformation( \Magento\Sales\Model\Order\Payment\Transaction::RAW_DETAILS );

		if ( ! isset( $paymentAdditionalInfo['transactionId'] ) && ! isset( $paymentAdditionalInfo['transaction_id'] ) ) {
			throw new Exception( 'The order was not properly complete. The required parameter Transaction ID is missing from order information. Online refund is not possible. Please contact customer support.' );
		}

		$transactionId = ( ! empty( $paymentAdditionalInfo['transactionId'] ) ) ? $paymentAdditionalInfo['transactionId'] : $paymentAdditionalInfo['transaction_id'];
		$refund        = $this->getRefundParams( $amount, $reason, $order, $transactionId );

		try {
			$response = $this->_client->postRefund( $refund, $transactionId );
		} catch ( \VerifoneEcommApi\ApiWrapper\Http\HttpException $exception ) {
			$this->_logger->critical( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
				json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

			return $this->failure();
		}

		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->_logger->critical( 'Error message', [
				'exception' => sprintf( 'There was an error decoding the json response from the api response, last known error was "%s", in file %s, at line %s. API response: %s',
					json_last_error(), __FILE__, __LINE__, $response )
			] );
			throw new Exception( 'There was an error in the application' );
		}
		if ( ! isset( $response['id'] )
		     || ! isset( $response['status'] )
		     || 'REFUNDED' !== $response['status']
		) {
			$this->_logger->critical( 'Error message', [ 'exception' => ' Refund Error: Unable to refund transaction.' ] );
			throw new Exception( 'Refund Error: Unable to refund transaction.' );
		}
		$payment->setAdditionalInformation(
			[ \Magento\Sales\Model\Order\Payment\Transaction::RAW_DETAILS => (array) $response ]
		);
		$payment->save();
		$order->setState( Order::STATE_CLOSED )->setStatus( $order->getConfig()->getStateDefaultStatus( Order::STATE_CLOSED ) );
		$order->save();
	}

	/**
	 * @param $amount
	 * @param $reason
	 * @param $order
	 * @param $transactionId
	 *
	 * @return array
	 */
	public function getRefundParams( $amount, $reason, $order, $transactionId ) {
		$date = new \DateTime( 'now' );

		return [
			// the api requires that all params are string...for this endpoint
			// in the initial call the amount has to be an int
			// but it seems to work with a string too
			// talk about consistency...
			'amount'            => (float) $amount * 100,
			'reason'            => $reason,
			'id'                => "{$transactionId}",
			'reference_id'      => "{$order->getQuoteId()}",
			'created_date_time' => "{$date->format( 'Y-m-d\TH:i:s\Z' )}"
		];
	}

	/**
	 * @param $quote
	 * @param $settings
	 * @param $paymentMethod
	 * @param $currency
	 * @param $customerId
	 * @param $locale
	 *
	 * @return array|\array[][]
	 */
	public function getCheckoutParams( $quote, $paymentMethod, $currency, $customerId, $locale ) {
		$customer_data = [];
		if ( $customerId !== '' ) {
			$customer_data['customer'] = $customerId;
		}

		$configurations = [
			'configurations' => [
				'card' => [
					'capture_now'         => ! $this->_verifoneSettings->getManualCapture(),
					'payment_contract_id' => $this->_verifoneSettings->getPaymentContractId(),
				],
			],
		];


		if ( $this->_verifoneSettings->is3dsEnabled() ) {
			$configurations['configurations']['card']['threedSecure'] = [
				'threeds_contract_id' => $this->_verifoneSettings->get3dsContractId(),
				'enabled'             => $this->_verifoneSettings->is3dsEnabled(),
				'transaction_mode'    => $this->_verifoneSettings->getDeviceType(),
			];
		}

		if ( $this->_verifoneSettings->getIsGooglePayEnabled() ) {
			$configurations['configurations']['google_pay'] = [
				'card'        => [
					'sca_compliance_level' => $this->_verifoneSettings->getGooglePayScaComplianceLevel(),

					'payment_contract_id' => $this->_verifoneSettings->getGooglePayContractId(),
				],
				'capture_now' => ! $this->_verifoneSettings->getManualCapture(),
			];

			if ( $this->_verifoneSettings->getGooglePay3dsEnable() ) {
				$configurations['configurations']['google_pay']['threed_secure'] = [
					'threeds_contract_id' => $this->_verifoneSettings->getGooglePayThreedsContractId(),
					'transaction_mode'    => 'S',
				];
			}
		}

		if ( $this->_verifoneSettings->getIsApplePayEnabled() ) {
			$configurations['configurations']['apple_pay'] = [
				'card'        => [
					'sca_compliance_level' => $this->_verifoneSettings->getApplePayScaComplianceLevel(),
					'payment_contract_id'  => $this->_verifoneSettings->getApplePayContractId(),
				],
				'capture_now' => ! $this->_verifoneSettings->getManualCapture(),
			];

			if ( $this->_verifoneSettings->getApplePay3dsEnable() ) {
				$configurations['configurations']['apple_pay']['threed_secure'] = [
					'threeds_contract_id' => $this->_verifoneSettings->getApplePayThreedsContractId(),
					'transaction_mode'    => 'S',
				];
			}
		}

		if ( $this->_verifoneSettings->getIsPaypalEnabled() ) {
			$configurations['configurations']['paypal'] = [
				'capture_now'         => ! $this->_verifoneSettings->getManualCapture(),
				'payment_contract_id' => $this->_verifoneSettings->getPaypalContractId(),
			];
		}

		// this is by far the most dumbest way to convert to cents
		// god knows what's gonna happen with stuff like YEN
		$amount = str_replace( [ ',', '.' ], '', $quote->getGrandTotal() * 100 );
		$locale = substr( $locale, 0, 2 );
		if ( in_array( $locale, self::NORWAY_CODES ) ) {
			$locale = "no";
		}
		$params = [
			'entity_id'          => $this->_verifoneSettings->getEntityId(),
			'currency_code'      => strtoupper( $currency ),
			'amount'             => $amount,
			'return_url'         => $paymentMethod->getReturnUrl() . '?refNo=' . $quote->getId(),
			'merchant_reference' => sprintf( '%s', $quote->getId() ),

			'i18n' => [
				// using substr is the recommended way to get the language
				// also silently fallback to en if the language is not supported
				'default_language' => in_array(
					$locale,
					self::SUPPORTED_I18N_LANGUAGES
				)
					? $locale
					: 'en'
				,
			],
		];

		$checkoutData = $params;
		$checkoutData = array_merge( $checkoutData, $customer_data );

		return array_merge( $checkoutData, $configurations );
	}


	/**
	 * @param $billingAddress
	 *
	 * @return array
	 */
	public function getBillingArray( $billingAddress ) {
		$billingMainArr = [
			'email_address' => substr( $billingAddress->getEmail(), 0, 250 ),
			'phone_number'  => preg_replace( '/[^0-9]+/', '', substr( $billingAddress->getTelephone(), 0, 20 ) ),
		];
		if ( ! empty( $billingAddress->getCompany() ) ) {
			$billingMainArr['company_name'] = substr( $billingAddress->getCompany(), 0, 100 );
		}

		$billingArr                = [
			'address_1'    => substr( $billingAddress->getStreet()[0], 0, 40 ),
			'address_2'    => ( ! empty( $billingAddress->getStreet()[1] ) ) ? substr( $billingAddress->getStreet()[1], 0, 40 ) : substr( $billingAddress->getStreet()[0], 0, 40 ),
			'city'         => substr( $billingAddress->getCity(), 0, 50 ),
			'country_code' => strtoupper( $billingAddress->getCountryId() ),
			'first_name'   => substr( $billingAddress->getFirstname(), 0, 50 ),
			'last_name'    => substr( $billingAddress->getLastname(), 0, 50 ),
			'phone'        => preg_replace( '/[^0-9]+/', '', substr( $billingAddress->getTelephone(), 0, 20 ) ),
			'postal_code'  => substr( $billingAddress->getPostcode(), 0, 10 ),
			'state'        => ( ! empty( $billingAddress->getRegion() ) ) ? substr( $billingAddress->getRegion(), 0, 10 ) : 'XX'
		];
		$billingMainArr['billing'] = $billingArr;

		return $billingMainArr;
	}

	/**
	 * @param $shippingAddress
	 *
	 * @return array
	 */
	public function getShippingArray( $shippingAddress, $billingAddress ) {
		$billingState                = ( ! empty( $billingAddress->getRegion() ) ) ? substr( $billingAddress->getRegion(), 0, 10 ) : 'XX';
		$shippingArr                 = [
			'address_1'    => ( ! empty( $shippingAddress->getStreet()[0] ) ) ? substr( $shippingAddress->getStreet()[0], 0, 40 ) : substr( $billingAddress->getStreet()[0], 0, 40 ),
			'address_2'    => ( ! empty( $shippingAddress->getStreet()[1] ) ) ? substr( $shippingAddress->getStreet()[1], 0, 40 ) : substr( $billingAddress->getStreet()[0], 0, 40 ),
			'city'         => ( ! empty( $shippingAddress->getCity() ) ) ? substr( $shippingAddress->getCity(), 0, 50 ) : substr( $billingAddress->getCity(), 0, 50 ),
			'country_code' => ( ! empty( $shippingAddress->getCountryId() ) ) ? strtoupper( $shippingAddress->getCountryId() ) : strtoupper( $billingAddress->getCountryId() ),
			'first_name'   => ( ! empty( $shippingAddress->getFirstname() ) ) ? substr( $shippingAddress->getFirstname(), 0, 50 ) : substr( $billingAddress->getFirstname(), 0, 50 ),
			'last_name'    => ( ! empty( $shippingAddress->getLastname() ) ) ? substr( $shippingAddress->getLastname(), 0, 50 ) : substr( $billingAddress->getLastname(), 0, 50 ),
			'phone'        => ( ! empty( $shippingAddress->getTelephone() ) ) ? preg_replace( '/[^0-9]+/', '', substr( $shippingAddress->getTelephone(), 0, 20 ) ) : preg_replace( '/[^0-9]+/', '', substr( $billingAddress->getTelephone(), 0, 20 ) ),
			'postal_code'  => ( ! empty( $shippingAddress->getPostcode() ) ) ? substr( $shippingAddress->getPostcode(), 0, 10 ) : substr( $billingAddress->getPostcode(), 0, 10 ),
			'state'        => ( ! empty( $shippingAddress->getRegion() ) ) ? substr( $shippingAddress->getRegion(), 0, 50 ) : $billingState,
		];
		$shippingMainArr['shipping'] = $shippingArr;

		return $shippingMainArr;
	}

	/**
	 * @return array
	 */
	private function failure() {
		return [
			'result'   => 'failure',
			'messages' => 'There has been an error processing your order',
		];
	}

	/**
	 * @return \Magento\Sales\Model\Service\InvoiceService
	 */
	public function getInvoiceService() {
		return $this->_invoiceService;
	}

	/**
	 * @return \Magento\Framework\DB\TransactionFactory
	 */
	public function getTransactionFactory() {
		return $this->_transactionFactory;
	}

	/**
	 * @param $order
	 * @param $payment
	 * @param $transactionId
	 * @param $result
	 */
	public function createTransactionForCapturePayment( $order, $payment, $transactionId, $result ) {
		$message     = sprintf( 'Order with transaction ID %s was authorized. Authorization ID %s', $transactionId, $result['id'] );
		$transaction = $this->_transactionBuilder->setPayment( $payment )
		                                         ->setOrder( $order )
		                                         ->setTransactionId( $transactionId )
		                                         ->setFailSafe( true )
		                                         ->build( \Magento\Sales\Model\Order\Payment\Transaction::TYPE_CAPTURE );
		$payment->addTransactionCommentsToOrder( $transaction, $message );
		$payment->save();
	}

	/**
	 * @param $order
	 * @param $payment
	 * @param $transactionId
	 * @param $result
	 */
	public function createTransactionForVoid( $order, $payment, $transactionId, $result ) {
		$message     = sprintf( 'Order with transaction ID %s was voided. Void ID %s', $transactionId, $result['id'] );
		$trans       = $this->_transactionBuilder;
		$transaction = $trans->setPayment( $payment )
		                     ->setOrder( $order )
		                     ->setTransactionId( $transactionId )
		                     ->setFailSafe( true )
		                     ->build( \Magento\Sales\Model\Order\Payment\Transaction::TYPE_VOID );

		$payment->addTransactionCommentsToOrder( $transaction, $message );
		$payment->save();
	}
}
